/*
 * Rewrites individual methods to contain checks
 */

package rewrite.rewriter;

import rewrite.pointcut.*;
import rewrite.util.*;

import java.io.*;
import java.util.*;
import java.util.jar.*;

import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.apache.bcel.util.*;
import org.apache.bcel.Repository;
import org.apache.bcel.Constants;
import org.apache.bcel.verifier.statics.*;
import org.apache.bcel.verifier.*;
import org.apache.bcel.verifier.structurals.*;

import spox.policy.*;
import spox.policy.pointcut.*;
import spox.policy.PolicyComponent.*;
import spox.policy.pointcut.PointcutComponent.Pointcut;

public class MethodRewriter
{
    // Fields
    private PolicyClass policyClass = null;
    private ClassGen cgen = null;
    private MethodGen mgen = null;
    private Policy policy = null;
    private HashMap<String, ClassGen> classMap = null;
    
    private ArrayList<Edge> edge = new ArrayList<Edge>();
    
    private boolean injectCode = true;
    
    private InstructionHandle commitHandle = null;
    
    // Accessors
    public MethodGen getMgen()
    { return (MethodGen)mgen.clone(); }

    // Constructor
    public MethodRewriter (
            PolicyClass policyClass,
            ClassGen cgen,
            MethodGen mgen,
            Policy policy,
            HashMap<String, ClassGen> classMap)
    {
        this.policyClass = policyClass;
        this.cgen = cgen;
        this.mgen = mgen;
        this.policy = policy;
        this.classMap = classMap;
    }
    
    // Methods
    public static MethodGen rewriteMethod(
            PolicyClass policyClass,
            ClassGen cgen,
            MethodGen mgen,
            Policy policy,
            HashMap<String, ClassGen> classMap,
            Method oldMethod)
        throws UnenforceableException, ClassNotFoundException
    {
        // This is just a convenience method to avoid having to explicitly
        // create an instance of MethodRewriter prior to rewriting a method
        MethodRewriter methodRewriter = new MethodRewriter(
                policyClass, cgen, mgen, policy, classMap);
        methodRewriter.rewriteMethod(oldMethod);
        return methodRewriter.getMgen();
    }

    public boolean rewriteMethod (Method oldMethod)
        throws UnenforceableException, ClassNotFoundException
    {
        // Whenever we find a command which matches some pointcut label
        // in a policy edge, we call the appropriate edge method
        // Returns true if checks were inserted occurred, false otherwise

        // Set up the construction tools
        InstructionFactory ifact = new InstructionFactory(cgen);
        InstructionList ilist = new InstructionList();
        Method method = mgen.getMethod();

        boolean retVal = false;

        edge = new ArrayList<Edge>(policy.getData().edge);
        ArrayList<Forall> forall = policy.getData().forall;
        PointcutMatcher pm;
        
        // Add in any edges contained within foralls
        for (int i = 0; i < forall.size(); i++)
        {
            edge.addAll(PolicyComponent.extractEdges(
                    forall.get(i)));
        }

        // We'll search through every method in the class looking
        // for method calls to see if they match up to any "call" pointcuts
        ilist = mgen.getInstructionList();
        if (ilist != null)  // Exceptions fire if we try to use a null
        {                   // instruction list!

            // Before we start iterating through the list, see if
            // there are any matching "Execution" elements in the policy
            InstructionHandle startHandle = ilist.getStart();
            for (int j = 0; j < edge.size(); j++)
            {
                pm = new PointcutMatcher(edge.get(j).label.pointcut);
                if (checkExecution(
                        ilist, ifact, pm, edge.get(j), startHandle))
                    retVal = true;
            }

            // Now do all the other checks
            if (checkInstructionList(ilist))
                retVal = true;

//            mgen.stripAttributes(true);
//            mgen.setMaxStack();
//            mgen.setMaxLocals();
//            cgen.replaceMethod(oldMethod, mgen.getMethod());
//            oldMethod = mgen.getMethod();
//            
//            for (int i = 0; i < cgen.getMethods().length; i++)
//                if (cgen.getMethodAt(i).getSignature().equals(mgen.getSignature()))
//                    mgen = new MethodGen(cgen.getMethodAt(i), cgen.getClassName(), cgen.getConstantPool());
            
            // Scan through and re-align any instance-state-oriented
            // objects to point to their corresponding wrappers
            if (repointToStatefuls(ilist, ifact))
                retVal = true;
            
            // Now update the method
            mgen.stripAttributes(true);
            mgen.setMaxStack();
            mgen.setMaxLocals();
            cgen.replaceMethod(oldMethod, mgen.getMethod());

            // Dispose of our instruction list
            ilist.dispose();
        
        }
        
        return retVal;
        
    }
    
    private boolean repointToStatefuls(
            InstructionList ilist,
            InstructionFactory ifact)
        throws UnenforceableException
    {
        // For any instruction which refers to an object that is tied to an
        // instance state value, realign the object reference to the "Stateful"
        // class corresponding to it
        
        ConstantPoolGen cpgen = cgen.getConstantPool();
        InstructionHandle ih = null;
        Instruction instruction = null;
                
        boolean retVal = false;
        
        String typeName;
        String arrayHolder;
        
        // Check superclass reference
        if (StateHolder.isStateful(cgen.getSuperclassName()))
            cgen.setSuperclassName(StateHolder.constructStatefulName(cgen.getSuperclassName()));
        
        // Check class fields (Might consider moving this out)
//        Field[] fields = cgen.getFields();
//        for (int i = 0; i < fields.length; i++)
//        {
//            typeName = fields[i].getType().toString();
//            arrayHolder = "";
//            if (typeName.contains("[]"))
//                arrayHolder = typeName.substring(typeName.indexOf("[]"));
//            typeName = typeName.replace("[]","");
//            if (StateHolder.isStateful(typeName))
//            {
//                FieldGen newField = new FieldGen(fields[i], cgen.getConstantPool());
//                newField.setType(Misc.constructType(
//                            StateHolder.constructStatefulName(typeName) +
//                            arrayHolder));
//                cgen.replaceField(fields[i], newField.getField());
//                retVal = true;
//            }
//        }
        
        // First check method parameters and return types
//        Type[] argType = mgen.getArgumentTypes();
//        for (int i = 0; i < argType.length; i++)
//        {
//            typeName = argType[i].toString();
//            arrayHolder = "";
//            if (typeName.contains("[]"))
//                arrayHolder = typeName.substring(typeName.indexOf("[]"));
//            typeName = typeName.replace("[]","");
//            if (StateHolder.isStateful(typeName))
//            {
//                mgen.setArgumentType(
//                        i, Misc.constructType(
//                            StateHolder.constructStatefulName(typeName) +
//                            arrayHolder));
//                retVal = true;
//            }
//        }
//        typeName = mgen.getReturnType().toString();
//        arrayHolder = "";
//        if (typeName.contains("[]"))
//            arrayHolder = typeName.substring(typeName.indexOf("[]"));
//        typeName = typeName.replace("[]","");
//        if (StateHolder.isStateful(typeName))
//        {
//            mgen.setReturnType(Misc.constructType(
//                            StateHolder.constructStatefulName(typeName) +
//                            arrayHolder));
//            retVal = true;
//        }

        // Modify local variables as necessary
//        {
//            System.out.println(mgen.getName());
//            LocalVariableGen[] lvgen = mgen.getLocalVariables();
//            for (int i = 0; i < lvgen.length; i++)
//                if (StateHolder.isStateful(lvgen[i].getType().toString()))
//                {
//                    lvgen[i].setType(Misc.constructType(StateHolder.
//                            constructStatefulName(lvgen[i].getType().toString())));
//                    mgen.addLocalVariable(lvgen[i].getName(), lvgen[i].getType(),
//                            lvgen[i].getIndex(), lvgen[i].getStart(), lvgen[i].getEnd());
//                }
//        }
        
        // Now change any relevant instructions
        Iterator<InstructionHandle> iter = ilist.iterator();
        while (iter.hasNext())
        {
            ih = iter.next();
            instruction = ih.getInstruction();
            if (instruction instanceof NEW)
            {
                // If needed, replace the NEW with the Stateful version
                // (It will get upcast from this point on)
                NEW newInstruction = (NEW)instruction;
                typeName = newInstruction.getLoadClassType(cpgen).getClassName();
                if (StateHolder.isStateful(typeName))
                {
                    if (!injectCode)
                        return true;
                    StateHolder stateful = StateHolder.
                            getStateful(typeName);
                    String statefulName = stateful.getCgen().getClassName();
                    ih.setInstruction(ifact.createNew(statefulName));
                    retVal = true;
                }
//                else if (StateHolder.getStatefulSuperclass(typeName) != null)
//                {
//                    try {
//                        StateHolder.createStatefulClass(
//                                typeName,
//                                StateHolder.getStatefulSuperclass(typeName));
//                    } catch (ClassNotFoundException ex) {
//                        throw new UnenforceableException("Class " + typeName + " not found in classpath.");
//                    }
//                    if (!injectCode)
//                        return true;
//                    StateHolder stateful = StateHolder.
//                            getStateful(typeName);
//                    String statefulName = stateful.getCgen().getClassName();
//                    ih.setInstruction(ifact.createNew(statefulName));
//                    retVal = true;
//                }
            }
            else if (instruction instanceof InvokeInstruction)
            {
                InvokeInstruction invoke = (InvokeInstruction)instruction;

                // Grab the class name and do any necessary conversion
                String className = invoke.getReferenceType(cpgen).toString();
                Type objType = Misc.constructType(className);

//                // First argument types
//                Type[] newArgTypes = invoke.getArgumentTypes(cpgen);
//                String retTypeName = invoke.getReturnType(cpgen).toString().replace("[]", "");
//                for (int i = 0; i < newArgTypes.length; i++)
//                {
//                    typeName = newArgTypes[i].toString();
//                    arrayHolder = "";
//                    if (typeName.contains("[]"))
//                        arrayHolder = typeName.substring(typeName.indexOf("[]"));
//                    typeName = typeName.replace("[]","");
//                    if (StateHolder.isStateful(typeName))
//                    {
//                        newArgTypes[i] = Misc.constructType(
//                            StateHolder.constructStatefulName(typeName) +
//                            arrayHolder);
//                        retVal = true;
//                    }
//                }
//                if (!classMap.containsKey(className))
//                {
////                    if (argRetCount > 1)
////                        throw new UnenforceableException("Cannot resolve wrapper for return value of " + invoke.getSignature(cpgen));
////                    else
////                    {
//                    // Pop off everything so we can resolve wrapper issues
//                    // (And if we want to restore a wrapper from the return,
//                    // we need to dup that wrapper at the bottom of the stack)
//                    if (!injectCode)
//                        return true;
//                    Type[] type = invoke.getArgumentTypes(cpgen);
//                    LocalVariableGen[] lvgen = new LocalVariableGen[type.length];
//                    for (int i = 0; i < type.length; i++)
//                    {
//                        if (StateHolder.isStateful(type[i].toString()))
//                        {
//                            lvgen[i] = mgen.addLocalVariable(
//                                    UniqueNameGen.findUniqueLocalVariableName(
//                                        mgen, "lv"),
//                                    Misc.constructType(
//                                        StateHolder.constructStatefulName(type[i].toString())),
//                                    null, null);
//                        }
//                        else
//                        {
//                            lvgen[i] = mgen.addLocalVariable(
//                                    UniqueNameGen.findUniqueLocalVariableName(
//                                        mgen, "lv"),
//                                    type[i], null, null);
//                        }
//                    }
//                    for (int i = lvgen.length-1; i >= 0; i--)
//                    {
//                        ilist.insert(ih, ifact.createStore(
//                            lvgen[i].getType(), lvgen[i].getIndex()));
//                    }
//                    LocalVariableGen virtualInvokeLvgen = null;
//                    if (invoke.getOpcode() != Constants.INVOKESTATIC)
//                    {
//                        virtualInvokeLvgen = mgen.addLocalVariable(
//                            UniqueNameGen.findUniqueLocalVariableName(mgen, "lv"),
//                            invoke.getReferenceType(cpgen), null, null);
//                        ilist.insert(ih, ifact.createStore(
//                                virtualInvokeLvgen.getType(),
//                                virtualInvokeLvgen.getIndex()));
//                    }
//                    if (invoke.getOpcode() != Constants.INVOKESTATIC)
//                    {
//                        ilist.insert(ih, ifact.createLoad(
//                                virtualInvokeLvgen.getType(),
//                                virtualInvokeLvgen.getIndex()));
//                    }
//                    for (int i = 0; i < lvgen.length; i++)
//                    {
//                        ilist.insert(ih, ifact.createLoad(
//                            lvgen[i].getType(), lvgen[i].getIndex()));
//                        if (StateHolder.isStateful(type[i].toString()))
//                            ilist.insert(ih, ifact.createGetField(
//                                StateHolder.constructStatefulName(type[i].toString()),
//                                StateHolder.WRAPPED_OBJ_NAME, type[i]));
//                    }
//                    retVal = true;
//                }
////                }
//                
//                // Second return type
//                Type newReturnType = invoke.getReturnType(cpgen);
//                typeName = newReturnType.toString();
//                arrayHolder = "";
//                if (typeName.contains("[]"))
//                    arrayHolder = typeName.substring(typeName.indexOf("[]"));
//                typeName = typeName.replace("[]","");
//                if (StateHolder.isStateful(typeName))
//                {
//                    if (classMap.containsKey(className))
//                    {
//                        // It's rewritable, so we can expect it to return the wrapper
//                        newReturnType = Misc.constructType(
//                                StateHolder.constructStatefulName(typeName) +
//                                arrayHolder);
//                        retVal = true;
//                    }
//                    else
//                    {
//                        // It's NOT rewritable, so we should not change the reference
//                        // However, we should handle the returned value somehow;
//                        // For now, we create a fresh new wrapper for it
//                        if (!injectCode)
//                            return true;
//                        LocalVariableGen retVarLvgen = mgen.addLocalVariable(
//                                UniqueNameGen.findUniqueLocalVariableName(mgen, "lv"),
//                                newReturnType, null, null);
//                        InstructionList appendIlist = new InstructionList();
//                        appendIlist.append(ifact.createStore(newReturnType,
//                                retVarLvgen.getIndex()));
//                        appendIlist.append(ifact.createNew(StateHolder.
//                                constructStatefulName(typeName)));
//                        appendIlist.append(ifact.DUP);
//                        appendIlist.append(ifact.createLoad(newReturnType,
//                                retVarLvgen.getIndex()));
//                        appendIlist.append(ifact.createInvoke(
//                                StateHolder.constructStatefulName(typeName),
//                                "<init>", Type.VOID, new Type[] {newReturnType},
//                                Constants.INVOKESPECIAL));
//                        ilist.append(ih, appendIlist);
//                        retVal = true;
//                    }
//                }

                // If this is a constructor call on a wrapped class, we need
                // to push the instance into the wrapper
                if (StateHolder.isStateful(className) &&
                        invoke.getMethodName(cpgen).equals("<init>"))
                {
                    if (!injectCode)
                        return true;
                    StateHolder stateful = StateHolder.
                            getStateful(className);
                    String statefulName = stateful.getCgen().getClassName();
                    ih.setInstruction(ifact.createInvoke(
                            statefulName, invoke.getMethodName(cpgen),
                            invoke.getReturnType(cpgen),
                            invoke.getArgumentTypes(cpgen),
                            invoke.getOpcode()));
                }
//                // We call the same class's method, but precede it with a
//                // stack reference conversion if non-static
//                else if (invoke.getOpcode() != Constants.INVOKESTATIC)
////                    invoke.getOpcode() != Constants.INVOKESTATIC_QUICK &&
////                    invoke.getOpcode() != Constants.INVOKENONVIRTUAL &&
////                    invoke.getOpcode() != Constants.INVOKENONVIRTUAL_QUICK)
//                {
//                    if (StateHolder.isStateful(className))
//                    {
//                        if (!injectCode)
//                            return true;
//                        // Pop everything off so we can alter the object reference
//                        Type[] type = invoke.getArgumentTypes(cpgen);
//                        LocalVariableGen[] lvgen = new LocalVariableGen[type.length];
//                        for (int i = 0; i < type.length; i++)
//                            lvgen[i] = mgen.addLocalVariable(
//                                    UniqueNameGen.findUniqueLocalVariableName(
//                                        mgen, "lv"),
//                                    type[i], null, null);
//                        InstructionHandle startIh =
//                                ilist.insert(ih, createStoreIlist(cgen, mgen, lvgen, type));
////                        if (!classMap.containsKey(className) &&
////                                invoke.getReturnType(cpgen).toString().replace("[]", "").equals(className))
////                            ilist.insert(ih, ifact.DUP); // Prepares for return value restoration
//                        ilist.insert(ih, ifact.createGetField(
//                                StateHolder.constructStatefulName(className),
//                                StateHolder.WRAPPED_OBJ_NAME, objType));
//                        ilist.insert(ih, createLoadIlist(cgen, mgen, lvgen, type));
////                        for (int i = 0; i < lvgen.length; i++)
////                            lvgen[i].setStart(startIh);
//                        retVal = true;
//                    }
//                }
//                if (className.contains("[]"))
//                    className = Misc.constructType(className).getSignature();
////                String newClassName;
////                String oldClassName = invoke.getReferenceType(cpgen).toString();
////                if (!InstanceStateHolder.isWrapped(oldClassName))
////                    newClassName = oldClassName;
////                else
////                {
////                    newClassName = InstanceStateHolder.constructWrappedName(oldClassName);
////                    retVal = true;
////                }
//                // Finally, replace the instruction
//                if (classMap.containsKey(className))
//                {
//                    ih.setInstruction(ifact.createInvoke(
//                            className,
//                            invoke.getMethodName(cpgen),
//                            newReturnType,
//                            newArgTypes,
//                            invoke.getOpcode()));
//                }
            }
//            else if (instruction instanceof GETFIELD ||
//                     instruction instanceof PUTFIELD)
//            {
//                FieldInstruction field = (FieldInstruction)instruction;
//                // First grab the class name and do any necessary conversion
//                String className = field.getReferenceType(cpgen).toString();
//                Type objType = Misc.constructType(className);
//                // First class name
//                if (StateHolder.isStateful(className))
//                {
//                    if (!injectCode)
//                        return true;
//                    // Alter the object reference
//                    if (field instanceof GETFIELD)
//                    {
//                        ilist.insert(ih, ifact.createGetField(
//                                StateHolder.constructStatefulName(className),
//                                StateHolder.WRAPPED_OBJ_NAME, objType));
//                        retVal = true;
//                    }
//                    else // PUTFIELD
//                    {
//                        Type[] type = new Type[] {field.getFieldType(cpgen)};
//                        LocalVariableGen[] lvgen = new LocalVariableGen[]
//                            {mgen.addLocalVariable(
//                                     UniqueNameGen.findUniqueLocalVariableName(
//                                        mgen, "lv"),
//                                    type[0], null, null)};
//                        InstructionHandle startIh =
//                                ilist.insert(ih, createStoreIlist(cgen, mgen, lvgen, type));
//                        ilist.insert(ih, ifact.createGetField(
//                                StateHolder.constructStatefulName(className),
//                                StateHolder.WRAPPED_OBJ_NAME, objType));
//                        ilist.insert(ih, createLoadIlist(cgen, mgen, lvgen, type));
//    //                        for (int i = 0; i < lvgen.length; i++)
//    //                            lvgen[i].setStart(startIh);
//                        retVal = true;
//                    }
//                }
//                if (className.contains("[]"))
//                    className = Misc.constructType(className).getSignature();
//                // Second field type
//                Type newFieldType = field.getFieldType(cpgen);
//                typeName = newFieldType.toString();
//                arrayHolder = "";
//                if (typeName.contains("[]"))
//                    arrayHolder = typeName.substring(typeName.indexOf("[]"));
//                typeName = typeName.replace("[]","");
//                if (StateHolder.isStateful(typeName))
//                {
//                    newFieldType = Misc.constructType(
//                            StateHolder.constructStatefulName(typeName) +
//                            arrayHolder);
//                    retVal = true;
//                }
//                // Finally, replace the instruction
//                ih.setInstruction(ifact.createFieldAccess(
//                        className,
//                        field.getFieldName(cpgen),
//                        newFieldType,
//                        field.getOpcode()));
//            }
//            else if (instruction instanceof CPInstruction)
//            {
//                CPInstruction cpInstruction = (CPInstruction)instruction;
//                Type newType = cpInstruction.getType(cpgen);
//                typeName = newType.toString();
//                arrayHolder = "";
//                if (typeName.contains("[]"))
//                    arrayHolder = typeName.substring(typeName.indexOf("[]"));
//                typeName = typeName.replace("[]","");
//                if (StateHolder.isStateful(typeName))
//                    cpInstruction.setIndex(cpgen.addClass((ObjectType)
//                        Misc.constructType(
//                            StateHolder.constructStatefulName(typeName) +
//                            arrayHolder)));
//            }
            //if (instruction instanceof org.apache.bcel.generic.)
        }
        
        return false;
    }
    
    public static InstructionList createStoreIlist(
            ClassGen cgen,
            MethodGen mgen,
            LocalVariableGen[] lvgen,
            Type[] type)
        throws UnenforceableException
    {
        // Creates an instruction list that pulls objects off the stack
        // so that we can do stuff to specific objects
        
        if (lvgen.length != type.length)
            throw new UnenforceableException(
                        "Incompatible type/local variable array lengths.");
        
        // Create local variable stores so we have places to put things
        InstructionFactory ifact = new InstructionFactory(cgen);
        InstructionList ilist = new InstructionList();
        
        for (int i = type.length-1; i >= 0; i--)
        {
            ilist.append(ifact.createStore(type[i], lvgen[i].getIndex()));
        }

        return ilist;
    }
    
    public static InstructionList createLoadIlist(
            ClassGen cgen,
            MethodGen mgen,
            LocalVariableGen[] lvgen,
            Type[] type)
        throws UnenforceableException
    {
        // Creates an instruction list that pushes objects back onto the stack
        // (meant to be used with createStoreIlist, above)
        
        if (lvgen.length != type.length)
            throw new UnenforceableException(
                        "Incompatible type/local variable array lengths.");
        
        // Create local variable stores so we have places to put things
        InstructionFactory ifact = new InstructionFactory(cgen);
        InstructionList ilist = new InstructionList();
        
        for (int i = 0; i < type.length; i++)
        {
            ilist.append(ifact.createLoad(type[i], lvgen[i].getIndex()));
        }
        
        return ilist;
    }

    public boolean checkInstructionList(InstructionList ilist)
        throws UnenforceableException, ClassNotFoundException
    {
        // Checks a list of instructions to see if they are policy relevant
        // If so, inserts checks as appropriate. Returns true if anything
        // was found.
        
        // Make sure we aren't looking at a null instruction list
        if (ilist == null)
            return false;

        InstructionFactory ifact = new InstructionFactory(cgen);
        ConstantPoolGen cpgen = cgen.getConstantPool();
        String cname = cgen.getClassName();
        Method[] methods = cgen.getMethods();
        Method method = mgen.getMethod();
        Iterator<InstructionHandle> iterator = null;
        Instruction instruction;
        InstructionHandle ih;
        short opcode;
        
        boolean checkPlaced;
        
        boolean retVal = false;

        Pointcut pointcut;

        iterator = ilist.iterator();
        while (iterator.hasNext())
        {
            
            ih = iterator.next();
            if (injectCode)
                commitHandle = ih;
            instruction = ih.getInstruction();
            opcode = instruction.getOpcode();
            ArrayList<InstructionList> commitIlistBefore = new ArrayList<InstructionList>();
            ArrayList<InstructionList> commitIlistAfter = new ArrayList<InstructionList>();
            for (int j = 0; j < edge.size(); j++)
            {
                checkPlaced = false;
                pointcut = edge.get(j).label.pointcut;
                PointcutMatcher pm = new PointcutMatcher(pointcut);

                // If it's a NEW, see if we need create a new Stateful class
                if (instruction instanceof NEW)
                {
                    String typeName = ((NEW)instruction).getLoadClassType(cpgen).toString();
                    if (!StateHolder.isStateful(typeName))
                    {
                        if (StateHolder.getStatefulSuperclass(typeName) != null)
                        {
                            try {
                                StateHolder.createStatefulClass(
                                        typeName,
                                        StateHolder.getStatefulSuperclass(typeName));
                            } catch (ClassNotFoundException ex) {
                                throw new UnenforceableException("Class " + typeName + " not found in classpath.");
                            }
                        }
                    }
                }
                
                // See if the command is an invoke (method call)
                // of some kind
                if (instruction instanceof InvokeInstruction)
                {
                    InstructionList commitIlist = new InstructionList();
                    checkPlaced = checkCall(
                            ih, instruction,
                            ilist, ifact, pm, edge.get(j),
                            cpgen, commitIlist);
                    if (!commitIlist.isEmpty())
                    {
                        if (edge.get(j).after)
                            commitIlistAfter.add(commitIlist);
                        else
                            commitIlistBefore.add(commitIlist);
                    }
                }
                // Is it a field instruction then?
                else if (instruction instanceof FieldInstruction)
                {
                    InstructionList commitIlist = new InstructionList();
                    checkPlaced = checkFieldOp(
                            ih, instruction,
                            ilist, ifact, pm, edge.get(j),
                            cpgen, commitIlist);
                    if (!commitIlist.isEmpty())
                    {
                        if (edge.get(j).after)
                            commitIlistAfter.add(commitIlist);
                        else
                            commitIlistBefore.add(commitIlist);
                    }
                }
                // See if the instruction matches an instr pointcut
                if (!checkPlaced)
                {
                    InstructionList commitIlist = new InstructionList();
                    checkPlaced = checkInstr(
                        ih, instruction,
                        ilist, ifact, pm, edge.get(j),
                        cpgen, commitIlist);
                    if (!commitIlist.isEmpty())
                    {
                        if (edge.get(j).after)
                            commitIlistAfter.add(commitIlist);
                        else
                            commitIlistBefore.add(commitIlist);
                    }
                }
                if (checkPlaced)
                    retVal = true;
            }

            // Figure out where to place the commits
            for (int j = 0; j < commitIlistBefore.size(); j++)
                injectIlist(ilist, ih, commitIlistBefore.get(j), false);
            for (int j = 0; j < commitIlistAfter.size(); j++)
                injectIlist(ilist, commitHandle, commitIlistAfter.get(j), true);
//            injectIlist(ilist, commitHandle, commitIlist, true);

        }
        
        return retVal;
    }

    public boolean checkCall(
            InstructionHandle ih,
            Instruction instruction,
            InstructionList ilist,
            InstructionFactory ifact,
            PointcutMatcher pointcut,
            Edge edge,
            ConstantPoolGen cpgen,
            InstructionList commitIlist) // Note that commitIlist is output
        throws UnenforceableException, ClassNotFoundException
    {
        
        // Locals
        String methodName;
        Type classId;
        Type[] param;
        Type returnType;
        String fullMethodCall;
        InstructionHandle nextIh; // For branch redirects
        boolean returnValue = false;
        PointcutData pData = new PointcutData();
        Method method = mgen.getMethod();
        
        InvokeInstruction invokeInstruction = (InvokeInstruction)instruction;
        
        // Now find out what we're invoking
        fullMethodCall =
                invokeInstruction.getReferenceType(cpgen).toString() +
                "." + invokeInstruction.getMethodName(cpgen).toString();
        pData.instruction = instruction;
        pData.ih = ih;
        pData.cgen = cgen;
        pData.mgen = mgen;
        pData.method = method;
        pData.cpgen = cpgen;
        pData.classType =
                invokeInstruction.
                getReferenceType(cpgen);
        pData.memberName =
                invokeInstruction.
                getMethodName(cpgen).toString();
        pData.argType = invokeInstruction.
                getArgumentTypes(cpgen);
        pData.retType =
                invokeInstruction.
                getReturnType(cpgen);
        if (Misc.classExists(invokeInstruction.getReferenceType(cpgen).toString()))
        {
            Method calledMethod = Misc.getMethod(Repository.lookupClass(
                    invokeInstruction.getReferenceType(cpgen).toString()),
                    invokeInstruction.getMethodName(cpgen) + invokeInstruction.getSignature(cpgen));
            if (calledMethod != null)
            {
                pData.modifiers = calledMethod.getModifiers();
                if (calledMethod.getExceptionTable() != null)
                    pData.throwsType = Misc.constructType(calledMethod.getExceptionTable().getExceptionNames());
            }
        }
        pData.generateCode = injectCode;
        
        // Set up the variable-to-object name mapping
        for (int i = 0; i < edge.nodes.size(); i++)
        {
            if (!edge.nodes.get(i).objName.isEmpty())
            {
                pData.nodeObj.put(edge.nodes.get(i).objName,
                        Policy.getStateObject(
                            edge.nodes.get(i), policy.getData().state));
            }
        }

        // Store all the state objects for pointcut reference
        for (int i = 0; i < policy.getData().state.size(); i++)
            pData.stateObj.add(policy.getData().state.get(i).objName);

        // Now check to see if the comparison holds
        if (pointcut.matches(pData))
        {
            if (!injectCode)
                return true;
            
            InstructionList beforeIlist = new InstructionList(); // Goes before each edge
            InstructionList afterIlist = new InstructionList();
            InstructionList paramStoreIlist = new InstructionList(); // Stores instance state parameters
            InstructionList paramGetIlist = new InstructionList();  // Obtains instance state parameters
            // Get any instance parameters
            LocalVariableGen[] argLvgen = {};
            Type[] argType = new Type[pData.nodeObjIndex.values().size()];
            if (argType.length > 0)
            {
                // Is this an illegal constructor target check?
                if (((InvokeInstruction)instruction).getMethodName(cpgen).equals("<init>") &&
                        !edge.after)
                {
                    Iterator<Integer> iter = pData.nodeObjIndex.values().iterator();
                    while (iter.hasNext())
                    {
                        if (iter.next().intValue() == 0)
                            throw new UnenforceableException(
                                    "Cannot inject 'before' checks for instance targets tied to constructor calls.");
                    }
                }
                
                // Pop ALL the arguments off the stack so we can get what we need
                InvokeInstruction invoke = (InvokeInstruction)instruction;
                Type[] invokeArgType = null;
                if (invoke.getOpcode() == Constants.INVOKESTATIC)
                    invokeArgType = invoke.getArgumentTypes(cpgen);
                else
                {
                    invokeArgType = new Type[invoke.getArgumentTypes(cpgen).length+1];
                    invokeArgType[0] = invoke.getReferenceType(cpgen);
                    System.arraycopy(invoke.getArgumentTypes(cpgen), 0, invokeArgType, 1, invokeArgType.length-1);
                }
                argLvgen = new LocalVariableGen[invokeArgType.length];
                for (int i = 0; i < argLvgen.length; i++)
                    argLvgen[i] = mgen.addLocalVariable(UniqueNameGen.
                            findUniqueLocalVariableName(mgen, "lv"),
                            invokeArgType[i], null, null);
                paramStoreIlist.append(createStoreIlist(cgen, mgen, argLvgen, invokeArgType));
                paramStoreIlist.append(createLoadIlist(cgen, mgen, argLvgen, invokeArgType));
            }
            
            // If we're placing after, create and store a boolean so we know
            // what to do after the target instruction
            int boolCheckIndex = -1;
            if (edge.after)
            {
                LocalVariableGen boolCheckLvgen = mgen.addLocalVariable(
                        UniqueNameGen.findUniqueLocalVariableName(mgen, "lv"),
                        Type.BOOLEAN, null, null);
                boolCheckIndex = boolCheckLvgen.getIndex();
                // Note that we have to go in reverse order when inserting at the top
                pData.returnIlist.insert(ifact.createStore(Type.BOOLEAN, boolCheckIndex));
                pData.returnIlist.insert(ifact.ICONST_0);
                // Set the boolean to true at the end (for if the dynamic check
                // didn't give us a false value)
                beforeIlist.append(ifact.ICONST_1);
                beforeIlist.append(ifact.createStore(Type.BOOLEAN, boolCheckIndex));
            }
            
            // Load the appropriate object references
            int argValCurIndex = 0;
            for (int i = 0; i < edge.nodes.size(); i++)
            {
                String nodeObjName = edge.nodes.get(i).objName;
                if (!nodeObjName.isEmpty())
                {
                    if (pData.nodeObjIndex.containsKey(nodeObjName))
                    {
                        // Load up the appropriate argument
                        int lvIndex = pData.nodeObjIndex.get(
                                edge.nodes.get(i).objName);
                        argType[argValCurIndex] = argLvgen[lvIndex].getType();
                        if (lvIndex==0 && edge.after && pData.memberName.equals("<init>"))
                            paramGetIlist.insert(ifact.DUP); // Special case to handle constructors
                        else
                            paramGetIlist.append(ifact.createLoad(argType[argValCurIndex], argLvgen[lvIndex].getIndex()));
                        if (StateHolder.isStateful(argType[argValCurIndex].toString()))
                        {
                            paramGetIlist.append(ifact.createCheckCast((ReferenceType)Misc.constructType(
                                    StateHolder.constructInterfaceName(argType[argValCurIndex].toString()))));
                            argType[argValCurIndex] = Misc.constructType(
                                    StateHolder.constructInterfaceName(
                                        argType[argValCurIndex].toString()));
                        }
                        argValCurIndex++;
                    }
                    else if (pData.thisObj.containsKey(nodeObjName))
                    {
                        // Load the 'this' pointer (local variable index 0)
                        paramGetIlist.append(ifact.ALOAD_0);
                    }
                }
            }

            afterIlist.append(paramGetIlist.copy());
            afterIlist.append(ifact.createInvoke(policyClass.getClassName(),
                    policyClass.getEdgeMethodName(edge),
                    Type.VOID, policyClass.getEdgeMethodArgumentTypes(edge),
                    Constants.INVOKESTATIC));

            // Construct the commit
            commitIlist.append(paramGetIlist);
            commitIlist.append(ifact.createInvoke(policyClass.getClassName(),
                    policyClass.getCommitMethodName(edge),
                    Type.VOID, policyClass.getEdgeMethodArgumentTypes(edge),
                    Constants.INVOKESTATIC));
            
            // If we're placing afterIlist after the target instruction,
            // (i.e., if the edge is an "after" edge)
            // we need to check the boolean we set earlier to see if the
            // dynamic checks held up before calling the edge method
            if (edge.after)
            {
                InstructionHandle falseIh = afterIlist.append(ifact.NOP);
                afterIlist.insert(ifact.createBranchInstruction(Constants.IFEQ, falseIh));
                afterIlist.insert(ifact.createLoad(Type.BOOLEAN, boolCheckIndex));
            }
            
            // Grab the handle for the end of the afterIlist (for the commit)
            InstructionHandle afterIlistEnd = afterIlist.getEnd();
            
            // In case pData was modified to include a dynamic check,
            // we insert the edge invokestatic call immediately before
            // the last command of that dynamic check (which is a NOP)
            // Any false branches in the dynamic check will skip to the NOP,
            // thus ignoring our edge call (or the true value setting)
            pData.returnIlist.insert(paramStoreIlist);
            pData.returnIlist.insert(pData.returnIlist.getEnd(),
                    beforeIlist);
            if (!edge.after)
                pData.returnIlist.insert(pData.returnIlist.getEnd(), afterIlist);
            else
                injectIlist(ilist, ih, afterIlist, true);
            nextIh = injectIlist(ilist, ih,
                    pData.returnIlist, false);
            
            ilist.setPositions();
            if (afterIlistEnd.getPosition() > commitHandle.getPosition())
                commitHandle = afterIlistEnd;

            // Realign any branches so our check won't be
            // bypassed by a clever goto
            ilist.redirectBranches(ih, nextIh);
            ilist.redirectExceptionHandlers(mgen.getExceptionHandlers(), ih, nextIh);

            // Tell the user that we found something
            System.out.println(instruction.getName() + " " +
                    fullMethodCall + '\n' +
                    " found at offset " + ih.getPosition() +
                    " in Method " + method.getName() +
                    " in Class " + cgen.getClassName());

            returnValue = true;
        }
        
        return returnValue;
        
    }
    
    public boolean checkFieldOp(
            InstructionHandle ih,
            Instruction instruction,
            InstructionList ilist,
            InstructionFactory ifact,
            PointcutMatcher pointcut,
            Edge edge,
            ConstantPoolGen cpgen,
            InstructionList commitIlist) // CommitIlist is output
        throws UnenforceableException, ClassNotFoundException
    {
        
        // Locals
        String fieldName;
        Type classId;
        Type fieldType;
        String fullFieldName;
        InstructionHandle nextIh; // For branch redirects
        PointcutComponent.FieldOp.GetOrSet getOrSet;
        boolean returnValue = false;
        Method method = mgen.getMethod();

        PointcutData pData = new PointcutData();
        
        // Now find out what we're invoking
        fullFieldName =
                ((FieldInstruction)instruction).getReferenceType(cpgen).toString() +
                "." + ((FieldInstruction)instruction).getFieldName(cpgen).toString();
        pData.instruction = instruction;
        pData.ih = ih;
        pData.cgen = cgen;
        pData.mgen = mgen;
        pData.method = method;
        pData.cpgen = cpgen;
        pData.classType =
                ((FieldInstruction)instruction).
                getReferenceType(cpgen);
        pData.memberName =
                ((FieldInstruction)instruction).
                getFieldName(cpgen).toString();
        pData.retType =
                ((FieldInstruction)instruction).getFieldType(cpgen);
        if (Misc.classExists(pData.classType.toString()))
        {
            FieldInstruction fieldInstruction = (FieldInstruction)instruction;
            Field calledField = Misc.getField(Repository.lookupClass(
                    fieldInstruction.getReferenceType(cpgen).toString()),
                    fieldInstruction.getFieldName(cpgen) + fieldInstruction.getSignature(cpgen));
            if (calledField != null)
                pData.modifiers = calledField.getModifiers();
        }
        pData.generateCode = injectCode;
        
        // Set up the variable-to-object name mapping
        for (int i = 0; i < edge.nodes.size(); i++)
        {
            if (!edge.nodes.get(i).objName.isEmpty())
            {
                pData.nodeObj.put(edge.nodes.get(i).objName,
                        Policy.getStateObject(
                            edge.nodes.get(i), policy.getData().state));
            }
        }

        // Store all the state objects for pointcut reference
        for (int i = 0; i < policy.getData().state.size(); i++)
            pData.stateObj.add(policy.getData().state.get(i).objName);

        // Now check to see if the comparison holds
        if (pointcut.matches(pData))
        {
            if (!injectCode)
                return true;
            
            InstructionList beforeIlist = new InstructionList();
            InstructionList afterIlist = new InstructionList();
            // Get any instance parameters
            LocalVariableGen[] argLvgen = {};
            Type[] argType = new Type[pData.nodeObjIndex.values().size()];
            if (argType.length > 0)
            {
                // Pop off a setter argument if necessary
                FieldInstruction fieldInstruction = (FieldInstruction)instruction;
                Type[] fieldArgType = null;
                if (fieldInstruction.getOpcode() == Constants.PUTSTATIC)
                    fieldArgType = new Type[] {fieldInstruction.getFieldType(cpgen)};
                else if (fieldInstruction.getOpcode() == Constants.PUTFIELD)
                    fieldArgType = new Type[] {
                        fieldInstruction.getReferenceType(cpgen),
                        fieldInstruction.getFieldType(cpgen)};
                else if (fieldInstruction.getOpcode() == Constants.GETFIELD)
                    fieldArgType = new Type[] {fieldInstruction.getReferenceType(cpgen)};
                else
                    fieldArgType = new Type[0];
                argLvgen = new LocalVariableGen[fieldArgType.length];
                for (int i = 0; i < argLvgen.length; i++)
                    argLvgen[i] = mgen.addLocalVariable(UniqueNameGen.
                            findUniqueLocalVariableName(mgen, "lv"),
                            fieldArgType[i], null, null);
                beforeIlist.append(createStoreIlist(cgen, mgen, argLvgen, fieldArgType));
                beforeIlist.append(createLoadIlist(cgen, mgen, argLvgen, fieldArgType));
            }
            
            // If we're placing after, create and store a boolean so we know
            // what to do after the target instruction
            int boolCheckIndex = -1;
            if (edge.after)
            {
                LocalVariableGen boolCheckLvgen = mgen.addLocalVariable(
                        UniqueNameGen.findUniqueLocalVariableName(mgen, "lv"),
                        Type.BOOLEAN, null, null);
                boolCheckIndex = boolCheckLvgen.getIndex();
                // Note that we have to go in reverse order when inserting at the top
                pData.returnIlist.insert(ifact.createStore(Type.BOOLEAN, boolCheckIndex));
                pData.returnIlist.insert(ifact.ICONST_0);
                // Set the boolean to true at the end (for if the dynamic check
                // didn't give us a false value)
                beforeIlist.append(ifact.ICONST_1);
                beforeIlist.append(ifact.createStore(Type.BOOLEAN, boolCheckIndex));
            }
            
            // Load the appropriate object references
            int argValCurIndex = 0;
            for (int i = 0; i < edge.nodes.size(); i++)
            {
                String nodeObjName = edge.nodes.get(i).objName;
                if (!nodeObjName.isEmpty())
                {
                    if (pData.nodeObjIndex.containsKey(nodeObjName))
                    {
                        // Load up the appropriate argument
                        int lvIndex = pData.nodeObjIndex.get(
                                edge.nodes.get(i).objName);
                        argType[argValCurIndex] = argLvgen[lvIndex].getType();
                        afterIlist.append(ifact.createLoad(argType[argValCurIndex], argLvgen[lvIndex].getIndex()));
                        if (StateHolder.isStateful(argType[argValCurIndex].toString()))
                        {
                            afterIlist.append(ifact.createCheckCast((ReferenceType)Misc.constructType(
                                    StateHolder.constructInterfaceName(argType[argValCurIndex].toString()))));
//                                    StateHolder.constructInterfaceName(argType[curIndex].toString()))));
                            argType[argValCurIndex] = Misc.constructType(
                                    StateHolder.constructInterfaceName(
                                        argType[argValCurIndex].toString()));
                        }
                        argValCurIndex++;
                    }
                    else if (pData.thisObj.containsKey(nodeObjName))
                    {
                        // Load the 'this' pointer (local variable index 0)
                        afterIlist.append(ifact.ALOAD_0);
                    }
                }
            }
            afterIlist.append(ifact.createInvoke(policyClass.getClassName(),
                    policyClass.getEdgeMethodName(edge),
                    Type.VOID, policyClass.getEdgeMethodArgumentTypes(edge),
                    Constants.INVOKESTATIC));
            
            // TODO: Get field ops working properly with instance states

            // Construct the commit
            //commitIlist.append(paramIlist);
            commitIlist.append(ifact.createInvoke(policyClass.getClassName(),
                    policyClass.getCommitMethodName(edge),
                    Type.VOID, policyClass.getEdgeMethodArgumentTypes(edge),
                    Constants.INVOKESTATIC));

            // If we're placing afterIlist after the target instruction,
            // (i.e., if the edge is an "after" edge)
            // we need to check the boolean we set earlier to see if the
            // dynamic checks held up before calling the edge method
            if (edge.after)
            {
                InstructionHandle falseIh = afterIlist.append(ifact.NOP);
                afterIlist.insert(ifact.createBranchInstruction(Constants.IFEQ, falseIh));
                afterIlist.insert(ifact.createLoad(Type.BOOLEAN, boolCheckIndex));
            }
            
            // Grab the handle for the end of the afterIlist (for the commit)
            InstructionHandle afterIlistEnd = afterIlist.getEnd();

            // In case pData was modified to include a dynamic check,
            // we insert the edge invokestatic call immediately before
            // the last command of that dynamic check (which is a NOP)
            // Any false branches in the dynamic check will skip to the NOP,
            // thus ignoring our edge call (or the true value setting)
            pData.returnIlist.insert(pData.returnIlist.getEnd(),
                    beforeIlist);
            if (!edge.after)
                pData.returnIlist.insert(pData.returnIlist.getEnd(), afterIlist);
            else
                injectIlist(ilist, ih, afterIlist, true);
            nextIh = injectIlist(ilist, ih,
                    pData.returnIlist, false);

            ilist.setPositions();
            if (afterIlistEnd.getPosition() > commitHandle.getPosition())
                commitHandle = afterIlistEnd;

            // Realign any branches so our check won't be
            // bypassed by a clever goto
            ilist.redirectBranches(ih, nextIh);
            ilist.redirectExceptionHandlers(mgen.getExceptionHandlers(), ih, nextIh);

            // Tell the user that we found something
            System.out.println(instruction.getName() + " " +
                    fullFieldName + '\n' +
                    " found at offset " + ih.getPosition() +
                    " in Method " + method.getName() +
                    " in Class " + cgen.getClassName());

            returnValue = true;
        }
        
        return returnValue;
        
    }
    
    public boolean checkInstr(
            InstructionHandle ih,
            Instruction instruction,
            InstructionList ilist,
            InstructionFactory ifact,
            PointcutMatcher pointcut,
            Edge edge,
            ConstantPoolGen cpgen,
            InstructionList commitIlist) // commitIlist is output
        throws UnenforceableException, ClassNotFoundException
    {
        
        // Locals
        InstructionHandle nextIh; // For branch redirects
        boolean returnValue = false;
        
        PointcutData pData = new PointcutData();
        
        Method method = mgen.getMethod();

        pData.instruction = instruction;
        pData.ih = ih;
        pData.cgen = cgen;
        pData.mgen = mgen;
        pData.method = method;
        pData.cpgen = cpgen;
        pData.generateCode = injectCode;
        
        // Set up the variable-to-object name mapping
        for (int i = 0; i < edge.nodes.size(); i++)
        {
            if (!edge.nodes.get(i).objName.isEmpty())
            {
                pData.nodeObj.put(edge.nodes.get(i).objName,
                        Policy.getStateObject(
                            edge.nodes.get(i), policy.getData().state));
            }
        }

        // Store all the state objects for pointcut reference
        for (int i = 0; i < policy.getData().state.size(); i++)
            pData.stateObj.add(policy.getData().state.get(i).objName);

        // Now check to see if the comparison holds
        if (pointcut.matches(pData))
        {
            if (!injectCode)
                return true;
            
            InstructionList beforeIlist = new InstructionList();
            InstructionList afterIlist = new InstructionList();
            
            // If we're placing after, create and store a boolean so we know
            // what to do after the target instruction
            int boolCheckIndex = -1;
            if (edge.after)
            {
                LocalVariableGen boolCheckLvgen = mgen.addLocalVariable(
                        UniqueNameGen.findUniqueLocalVariableName(mgen, "lv"),
                        Type.BOOLEAN, null, null);
                boolCheckIndex = boolCheckLvgen.getIndex();
                // Note that we have to go in reverse order when inserting at the top
                pData.returnIlist.insert(ifact.createStore(Type.BOOLEAN, boolCheckIndex));
                pData.returnIlist.insert(ifact.ICONST_0);
                // Set the boolean to true at the end (for if the dynamic check
                // didn't give us a false value)
                beforeIlist.append(ifact.ICONST_1);
                beforeIlist.append(ifact.createStore(Type.BOOLEAN, boolCheckIndex));
            }
            
            for (int i = 0; i < edge.nodes.size(); i++)
            {
                String nodeObjName = edge.nodes.get(i).objName;
                if (!nodeObjName.isEmpty())
                {
                    if (pData.thisObj.containsKey(nodeObjName))
                    {
                        // Load the 'this' pointer (local variable index 0)
                        afterIlist.append(ifact.ALOAD_0);
                    }
                }
            }
            afterIlist.append(ifact.createInvoke(policyClass.getClassName(),
                    policyClass.getEdgeMethodName(edge),
                    Type.VOID, policyClass.getEdgeMethodArgumentTypes(edge),
                    Constants.INVOKESTATIC));

            // Construct the commit
            commitIlist.append(ifact.createInvoke(policyClass.getClassName(),
                    policyClass.getCommitMethodName(edge),
                    Type.VOID, policyClass.getEdgeMethodArgumentTypes(edge),
                    Constants.INVOKESTATIC));

            // If we're placing afterIlist after the target instruction,
            // (i.e., if the edge is an "after" edge)
            // we need to check the boolean we set earlier to see if the
            // dynamic checks held up before calling the edge method
            if (edge.after)
            {
                InstructionHandle falseIh = afterIlist.append(ifact.NOP);
                afterIlist.insert(ifact.createBranchInstruction(Constants.IFEQ, falseIh));
                afterIlist.insert(ifact.createLoad(Type.BOOLEAN, boolCheckIndex));
            }
            
            // Grab the handle for the end of the afterIlist (for the commit)
            InstructionHandle afterIlistEnd = afterIlist.getEnd();

            // In case pData was modified to include a dynamic check,
            // we insert the edge invokestatic call immediately before
            // the last command of that dynamic check (which is a NOP)
            // Any false branches in the dynamic check will skip to the NOP,
            // thus ignoring our edge call (or the true value setting)
            pData.returnIlist.insert(pData.returnIlist.getEnd(),
                    beforeIlist);
            if (!edge.after)
                pData.returnIlist.insert(pData.returnIlist.getEnd(), afterIlist);
            else
                injectIlist(ilist, ih, afterIlist, true);
            nextIh = injectIlist(ilist, ih,
                    pData.returnIlist, false);

            ilist.setPositions();
            if (afterIlistEnd.getPosition() > commitHandle.getPosition())
                commitHandle = afterIlistEnd;

            // Realign any branches so our check won't be
            // bypassed by a clever goto
            ilist.redirectBranches(ih, nextIh);
            ilist.redirectExceptionHandlers(mgen.getExceptionHandlers(), ih, nextIh);

            // Tell the user that we found something
            System.out.println(instruction.getName() +
                    " found at offset " + ih.getPosition() +
                    " in Method " + method.getName() +
                    " in Class " + cgen.getClassName());

            returnValue = true;
        }
        
        return returnValue;
        
    }

    public boolean checkExecution(
            InstructionList ilist,
            InstructionFactory ifact,
            PointcutMatcher pointcut,
            Edge edge,
            InstructionHandle startHandle)
        throws UnenforceableException, ClassNotFoundException
    {
        
        // Locals
        boolean returnValue = false;
        
        PointcutData pData = new PointcutData();
        
        Method method = mgen.getMethod();

        // Build the pointcut data
        pData.cgen = cgen;
        pData.cpgen = cgen.getConstantPool();
        pData.mgen = mgen;
        pData.method = method;
        pData.cpgen = cgen.getConstantPool();
        pData.classType =
                Type.getType("L" + cgen.getClassName().replace('.','/') + ";");
        pData.memberName = method.getName();
        pData.modifiers = method.getModifiers();
        pData.argType = method.getArgumentTypes();
        pData.retType = method.getReturnType();
        pData.generateCode = injectCode;

        // Set up the variable-to-object name mapping
        for (int i = 0; i < edge.nodes.size(); i++)
        {
            if (!edge.nodes.get(i).objName.isEmpty())
            {
                pData.nodeObj.put(edge.nodes.get(i).objName,
                        Policy.getStateObject(
                            edge.nodes.get(i), policy.getData().state));
            }
        }

        // Store all the state objects for pointcut reference
        for (int i = 0; i < policy.getData().state.size(); i++)
            pData.stateObj.add(policy.getData().state.get(i).objName);

        // Check to see if the comparison holds
        if (pointcut.matches(pData))
        {
            if (!injectCode)
                return true;
            
            // Currently we can't enforce after-edges on executions
            if (edge.after)
                throw new UnenforceableException(
                        "Cannot place edge check after execution.");
        
            InstructionList injectIlist = new InstructionList();
            InstructionList commitIlist = new InstructionList();
            InstructionList paramIlist = new InstructionList();
            // Get any instance parameters
            Type[] argType = mgen.getArgumentTypes();
//            LocalVariableGen[] argLvgen = {};
//            Type[] argType = new Type[pData.nodeObjIndex.values().size()];
//            if (argType.length > 0)
//            {
//                // Pop all the arguments off the stack so we can get what we need
//                InvokeInstruction invoke = (InvokeInstruction)instruction;
//                Type[] invokeArgType = null;
//                if (invoke.getOpcode() == Constants.INVOKESTATIC)
//                    invokeArgType = invoke.getArgumentTypes(cpgen);
//                else
//                {
//                    invokeArgType = new Type[invoke.getArgumentTypes(cpgen).length+1];
//                    invokeArgType[0] = invoke.getReferenceType(cpgen);
//                    System.arraycopy(invoke.getArgumentTypes(cpgen), 0, invokeArgType, 1, invokeArgType.length-1);
//                }
//                argLvgen = new LocalVariableGen[invokeArgType.length];
//                for (int i = 0; i < argLvgen.length; i++)
//                    argLvgen[i] = mgen.addLocalVariable(UniqueNameGen.
//                            findUniqueLocalVariableName(mgen, "lv"),
//                            invokeArgType[i], null, null);
//                injectIlist.append(createStoreIlist(cgen, mgen, argLvgen, invokeArgType));
//                injectIlist.append(createLoadIlist(cgen, mgen, argLvgen, invokeArgType));
//            }
            // Load the appropriate object references
            for (int i = 0; i < edge.nodes.size(); i++)
            {
                String nodeObjName = edge.nodes.get(i).objName;
                if (!nodeObjName.isEmpty())
                {
                    if (pData.nodeObjIndex.containsKey(nodeObjName))
                    {
                        // Load up the appropriate argument
                        int argIndex = pData.nodeObjIndex.get(
                                edge.nodes.get(i).objName);
                        int argOffset = 0;
                        if (!mgen.isStatic())
                            argOffset = 1;
                        paramIlist.append(ifact.createLoad(argType[argIndex],
                                argIndex+argOffset));
                        if (StateHolder.isStateful(argType[argIndex].toString()))
                        {
                            paramIlist.append(ifact.createCheckCast((ReferenceType)Misc.constructType(
                                    StateHolder.constructInterfaceName(argType[argIndex].toString()))));
                            argType[argIndex] = Misc.constructType(
                                    StateHolder.constructInterfaceName(
                                        argType[argIndex].toString()));
                        }
                    }
                    else if (pData.thisObj.containsKey(nodeObjName) && !mgen.isStatic())
                    {
                        // Load the 'this' pointer (local variable index 0)
                        paramIlist.append(ifact.ALOAD_0);
                    }
                }
            }
            injectIlist.append(paramIlist.copy());
            injectIlist.append(ifact.createInvoke(policyClass.getClassName(),
                    policyClass.getEdgeMethodName(edge),
                    Type.VOID, policyClass.getEdgeMethodArgumentTypes(edge),
                    Constants.INVOKESTATIC));
            
            commitIlist.append(paramIlist);
            commitIlist.append(ifact.createInvoke(policyClass.getClassName(),
                    policyClass.getCommitMethodName(edge),
                    Type.VOID, policyClass.getEdgeMethodArgumentTypes(edge),
                    Constants.INVOKESTATIC));
            
            // In case pData was modified to include a dynamic check,
            // we insert the edge invokestatic call immediately before
            // the last command of that dynamic check (which is a NOP)
            // Any false branches in the dynamic check will skip to the NOP,
            // thus ignoring our edge call
            pData.returnIlist.insert(pData.returnIlist.getEnd(),
                    injectIlist);

            // Call our edge function with an invokestatic
            injectIlist(ilist, ilist.getStart(),
                    pData.returnIlist, false);
            injectIlist(ilist, startHandle,
                    commitIlist, false);

            // Tell the user that we found something
            System.out.println("Method " + method.getName() + "\n" +
                    " execution found in Class " + cgen.getClassName());

            returnValue = true;
        }
        
        return returnValue;
    }
    
    private InstructionHandle injectIlist(
            InstructionList ilist,
            InstructionHandle insertHandle,
            InstructionList newIlist,
            boolean after)
        throws UnenforceableException, ClassNotFoundException
    {
        // Checks the newIlist first to see if it violates the policy.
        // If so, throws UnenforceableException, otherwise injects the code
        // into ilist and returns the appropriate InstructionHandle
        
        // Set injectCode to false so we don't recursively add instructions!
        injectCode = false;
        if (checkInstructionList(newIlist))
            throw new UnenforceableException("Cannot enforce policy on injected checks.");
        injectCode = true;

        InstructionHandle ih = null;
        if (!after)
            ih = ilist.insert(insertHandle, newIlist);
        else
            ih = ilist.append(insertHandle, newIlist);
                
        return ih;
    }

}
