/*
 * Handles a security-relevant class by either adding a state attribute to
 * the original class or, if that class is unavailable, creating an inherited,
 * "stateful" class that contains both a state attribute.
 */

package rewrite.rewriter;

import rewrite.util.*;

import java.io.*;
import java.util.*;
import java.util.jar.*;

import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.apache.bcel.util.*;
import org.apache.bcel.Constants;
import org.apache.bcel.Repository;

import spox.policy.*;
import spox.policy.pointcut.*;
import spox.policy.PolicyComponent.*;

public class StateHolder
{

    // Constants
    public static final String STATEFUL_NAME_PREFIX = "security.stateful.Stateful";
    public static final String INTERFACE_NAME_PREFIX = "security.state.";
    public static final String FIELD_NAME_PREFIX = "InstanceStateField_";
    public static final String WRAPPED_OBJ_NAME = "object";
    
    // Static fields
    private static HashMap<String, StateHolder> statefulMap =
            new HashMap<String, StateHolder>();
    private static HashMap<String, StateHolder> stateMap =
            new HashMap<String, StateHolder>();
    private static HashMap<String, ClassGen> interfaceMap =
            new HashMap<String, ClassGen>();
    private static HashMap<String, ClassGen> classMap =
            new HashMap<String, ClassGen>();
    
    private static Policy policy = new Policy();

    // Instance Fields
    private ClassGen cgen;
    private String oldClassName = null;
    
    // Constructor
    private StateHolder(ClassGen cgen, String oldClassName)
        throws UnenforceableException
    {
        this.cgen = cgen;
        this.oldClassName = oldClassName;
    }
    
    // Accessors
    public String getOldClassName()
    { return oldClassName; }
    
    public ClassGen getCgen()
    { return (ClassGen)cgen.clone(); }
    
    // Static methods
    public static void setClassMap(HashMap<String, ClassGen> classMap)
    { StateHolder.classMap = classMap; }
    
    public static void setPolicy(Policy policy)
    { StateHolder.policy = policy; }
    
    public static StateHolder getStateful(String className)
    { return statefulMap.get(className); }
    
    public static boolean isStateful(String className)
    { return statefulMap.containsKey(className); }

    /**
     * If the given class is a subclass of a stateful class, returns
     * the name of that class. If there are no stateful classes higher
     * in the hierarchy, returns null.
     * @param typeName - The name of the class to check.
     * @return The name of the closest stateful superclass, or null if
     * one does not exist.
     */
    public static String getStatefulSuperclass(String typeName)
    {
        try
        {
            JavaClass[] superclass = Repository.getSuperClasses(typeName);
            String retVal = null;
            for (int i = 0; (i < superclass.length && retVal==null); i++)
            {
                if (isStateful(superclass[i].getClassName()))
                    retVal = superclass[i].getClassName();
            }
            return retVal;
        }
        catch (ClassNotFoundException ex)
        {
            return null;
        }
    }
    
    /**
     * Creates a new stateful class for a class that does not carry its own
     * instance state, but is a subclass of some other state-carrying class.
     * 
     * @param className - The name of the class to make Stateful
     * @param statefulSuperclass - The name of the closest stateful superclass 
     * (can be retrieved with the getStatefulSuperclass method)
     */
    public static void createStatefulClass(
            String className,
            String statefulSuperclass)
        throws UnenforceableException, ClassNotFoundException
    {
        // Create a corresponding interface (with nothing in it)
        ClassGen interfaceCgen = interfaceMap.get(
            INTERFACE_NAME_PREFIX + className.replace(".", "_") + "State");
        if (interfaceCgen == null)
        {
            interfaceCgen = new ClassGen(
                    INTERFACE_NAME_PREFIX + className.replace(".", "_") + "State",
                    "java.lang.Object", null,
                    Constants.ACC_PUBLIC | Constants.ACC_INTERFACE,
                    new String[] {});
            interfaceCgen.addInterface(INTERFACE_NAME_PREFIX +
                    statefulSuperclass.replace(".", "_") + "State");
        }
        interfaceMap.put(interfaceCgen.getClassName(), interfaceCgen);

        // Create a Stateful class
        JavaClass jclass = Repository.lookupClass(className);
        if (jclass.isFinal())
            throw new UnenforceableException(
                    "Cannot create instance states for final class + " + className + ".");
        ClassGen scgen = getStateful(statefulSuperclass).getCgen();
        ClassGen cgen = new ClassGen(
                    STATEFUL_NAME_PREFIX + className.replace('.', '_'),
                    jclass.getClassName(),
                    null,
                    jclass.getModifiers(),
                    new String[] {interfaceCgen.getClassName()});
        ConstantPoolGen cpgen = cgen.getConstantPool();
        ArrayList<State> stateList = new ArrayList<State>();
        for (int i = 0; i < scgen.getFields().length; i++)
        {
            String fullname = scgen.getFields()[i].getName();
            stateList.add(Policy.getStateByName(
                    fullname.substring(FIELD_NAME_PREFIX.length()),
                    policy.getData().state));
        }
        
        StateHolder stateHolder = new StateHolder(cgen, className);
        stateHolder.createStatefulClass(jclass);
        stateHolder.addState(stateList);
        statefulMap.put(className, stateHolder);
    }

    public static String constructStatefulName(String className)
    {
        return STATEFUL_NAME_PREFIX + className.replace('.', '_');
    }
    
    public static String constructInterfaceName(String className)
    {
        return INTERFACE_NAME_PREFIX + className.replace(".","_") + "State";
    }
    
    public static void initialize()
    {
        statefulMap = new HashMap<String, StateHolder>();
        stateMap = new HashMap<String, StateHolder>();
        classMap = new HashMap<String, ClassGen>();
        interfaceMap = new HashMap<String, ClassGen>();
    }
    
    /**
     * Finalizes the stateful classes (and interfaces) by injecting their
     * ClassGens into the classMap.
     */
    public static void finalizeStatefuls()
    {
        
        Iterator<StateHolder> iter = statefulMap.values().iterator();
        while (iter.hasNext())
        {
            ClassGen cgen = iter.next().getCgen();
            classMap.put(cgen.getClassName(), cgen);
        }

        classMap.putAll(interfaceMap);
    }
    
    public static StateHolder injectState
            (String className, State state)
            throws UnenforceableException, ClassNotFoundException
    {
        // If the jar file contains the given class, add the state to it,
        // and create an interface to describe the necessary accessors
        
        // Note that for every explicitly generated Stateful..., we create
        // state-accessor interfaces for all subclasses that get passed in here
        // (along with in-between classes in the hierarchy)
        
        StateHolder stateHolder = null;
        
        ArrayList<State> stateList = new ArrayList<State>();
        stateList.add(state);
        
        if (classMap.containsKey(className))
        {
            // Class exists; add the field
            stateHolder =
                    new StateHolder(classMap.get(className), className);
            stateHolder.addState(stateList);
            stateMap.put(className, stateHolder);
        }
        else
        {
            if (!statefulMap.containsKey(className))
            {
                // First see what the top superclass of this class is in the
                // list of instance state parameters in the policy
                // (While we're at it, we grab the names of all states higher
                // in the hierarchy)
                JavaClass jclass = Repository.lookupClass(className);
                JavaClass[] superclassArray = jclass.getSuperClasses();
                State[] stateArray = policy.getData().state.toArray(new State[0]);
                int highestSuperclassIndex = -1;
                String highestSuperclassName = null;
                int lowestSuperclassIndex = superclassArray.length;
                String lowestSuperclassName = null;
                for (int i = 0; i < stateArray.length; i++)
                {
                    for (int j = 0; j < superclassArray.length; j++)
                    {
                        if (stateArray[i].objName.equals(superclassArray[j].getClassName()))
                        {
                            highestSuperclassIndex = Math.max(j, highestSuperclassIndex);
                            if (j == highestSuperclassIndex)
                                highestSuperclassName = stateArray[i].objName;
                            lowestSuperclassIndex = Math.min(j, lowestSuperclassIndex);
                            if (j == lowestSuperclassIndex)
                                lowestSuperclassName = stateArray[i].objName;
                            stateList.add(stateArray[i]);
                        }
                    }
                }

                // Create an interface ...State (unless one already exists)
                ClassGen interfaceCgen = interfaceMap.get(
                            INTERFACE_NAME_PREFIX + className.replace(".", "_") + "State");
                if (interfaceCgen == null)
                {
                    interfaceCgen = new ClassGen(
                            INTERFACE_NAME_PREFIX + className.replace(".", "_") + "State",
                            "java.lang.Object", null,
                            Constants.ACC_PUBLIC | Constants.ACC_INTERFACE,
                            new String[] {});
                    if (stateList.size() > 1) // There are interfaces above us
                        interfaceCgen.addInterface(INTERFACE_NAME_PREFIX +
                                lowestSuperclassName.replace(".", "_") + "State");
                }
                // Add the state accessors to the interface
                MethodGen interfaceMethod = new MethodGen(
                        Constants.ACC_PUBLIC | Constants.ACC_ABSTRACT,
                        Type.getType(Integer.class),
                        new Type[] {}, new String[] {},
                        "get" + FIELD_NAME_PREFIX + state.name,
                        interfaceCgen.getClassName(), new InstructionList(),
                        interfaceCgen.getConstantPool());
                interfaceMethod.removeLocalVariables();
                interfaceMethod.setInstructionList(null);
                interfaceCgen.addMethod(interfaceMethod.getMethod());
                interfaceMethod = new MethodGen(
                        Constants.ACC_PUBLIC | Constants.ACC_ABSTRACT,
                        Type.getType(Integer.class),
                        new Type[] {}, new String[] {},
                        "getTemp" + FIELD_NAME_PREFIX + state.name,
                        interfaceCgen.getClassName(), new InstructionList(),
                        interfaceCgen.getConstantPool());
                interfaceMethod.removeLocalVariables();
                interfaceMethod.setInstructionList(null);
                interfaceCgen.addMethod(interfaceMethod.getMethod());
                interfaceMethod = new MethodGen(
                        Constants.ACC_PUBLIC | Constants.ACC_ABSTRACT,
                        Type.VOID,
                        new Type[] {Type.getType(Integer.class)}, null,
                        "set" + FIELD_NAME_PREFIX + state.name,
                        interfaceCgen.getClassName(), new InstructionList(),
                        interfaceCgen.getConstantPool());
                interfaceMethod.removeLocalVariables();
                interfaceMethod.setInstructionList(null);
                interfaceCgen.addMethod(interfaceMethod.getMethod());
                interfaceMap.put(interfaceCgen.getClassName(), interfaceCgen);
                interfaceMethod = new MethodGen(
                        Constants.ACC_PUBLIC | Constants.ACC_ABSTRACT,
                        Type.VOID,
                        new Type[] {}, new String[] {},
                        "commit" + FIELD_NAME_PREFIX + state.name,
                        interfaceCgen.getClassName(), new InstructionList(),
                        interfaceCgen.getConstantPool());
                interfaceMethod.removeLocalVariables();
                interfaceMethod.setInstructionList(null);
                interfaceCgen.addMethod(interfaceMethod.getMethod());

                // Create a Stateful class
                if (jclass.isFinal())
                    throw new UnenforceableException(
                            "Cannot create instance states for final class + " + className + ".");
                ClassGen cgen = new ClassGen(
                            STATEFUL_NAME_PREFIX + className.replace('.', '_'),
                            jclass.getClassName(),
                            null,
                            jclass.getModifiers(),
                            new String[] {interfaceCgen.getClassName()});
                stateHolder = new StateHolder(cgen, className);
                stateHolder.createStatefulClass(jclass);
            }
            else
                stateHolder = statefulMap.get(className);
            stateHolder.addState(stateList);
            statefulMap.put(className, stateHolder);
        }
        
        return stateHolder;
        
    }

    // Instance methods
    private void createStatefulClass(JavaClass superclass)
        throws UnenforceableException, ClassNotFoundException
    {
        // Generates a stateful class corresponding to the current
        // classgen data
        
        InstructionList ilist = new InstructionList();
        ConstantPoolGen cpgen = cgen.getConstantPool();
        InstructionFactory ifact = new InstructionFactory(cpgen);
        
        // Build mirror constructors for every one in the superclass
        Method[] methodArray = superclass.getMethods();
        for (int i = 0; i < methodArray.length; i++)
        {
            if (methodArray[i].getName().equals("<init>"))
            {
                ilist = new InstructionList();
                ilist.append(ifact.ALOAD_0);
                for (int j = 0; j < methodArray[i].getArgumentTypes().length; j++)
                    ilist.append(ifact.createLoad(methodArray[i].getArgumentTypes()[j], j+1));
                ilist.append(ifact.createInvoke(superclass.getClassName(),
                        "<init>", methodArray[i].getReturnType(),
                        methodArray[i].getArgumentTypes(), Constants.INVOKESPECIAL));
                ilist.append(ifact.RETURN);
                MethodGen mg = new MethodGen(
                        methodArray[i].getAccessFlags(),
                        methodArray[i].getReturnType(),
                        methodArray[i].getArgumentTypes(),
                        null, "<init>", cgen.getClassName(),
                        ilist, cpgen);
                mg.setMaxLocals();
                mg.setMaxStack();
                mg.stripAttributes(true);
                cgen.addMethod(mg.getMethod());
                ilist.dispose();
            }
        }
    }
    
    public void addState(ArrayList<State> stateList)
        throws UnenforceableException
    {
        // Adds state fields and accessors (and temps) to the ClassGen
        
        // Set up the construction tools
        InstructionFactory ifact = new InstructionFactory(cgen);
        InstructionList ilist = new InstructionList();
        InstructionList initIlist = new InstructionList();
        ConstantPoolGen cpgen = cgen.getConstantPool();
        String cname = cgen.getClassName();
        MethodGen mg;
        Type IntegerType = Type.getType(java.lang.Integer.class);
        
        for (int i = 0; i < stateList.size(); i++)
        {
            State state = stateList.get(i);
        
            // Create the fields
            FieldGen fgen = new FieldGen(
                    Constants.ACC_PRIVATE,
                    IntegerType,
                    FIELD_NAME_PREFIX + state.name,
                    cpgen);
            cgen.addField(fgen.getField());
            FieldGen tempFgen = new FieldGen(
                    Constants.ACC_PRIVATE,
                    IntegerType,
                    "temp_" + FIELD_NAME_PREFIX + state.name,
                    cpgen);
            cgen.addField(tempFgen.getField());

            // Create a 'get' accessor
            InstructionList getIlist = new InstructionList();
            getIlist.append(ifact.ALOAD_0);
            getIlist.append(ifact.createGetField(cname, fgen.getName(), IntegerType));
            getIlist.append(ifact.createReturn(IntegerType));
            MethodGen mgen = new MethodGen(
                    Constants.ACC_PUBLIC, IntegerType,
                    new Type[] {}, new String[] {},
                    "get" + fgen.getName(),
                    cname, getIlist, cpgen);
            mgen.setMaxLocals();
            mgen.setMaxStack();
            mgen.stripAttributes(true);
            cgen.addMethod(mgen.getMethod());
            
            // Create a 'get' accessor for the 'temp' variable
            InstructionList getTempIlist = new InstructionList();
            getTempIlist.append(ifact.ALOAD_0);
            getTempIlist.append(ifact.createGetField(cname, "temp_"+fgen.getName(), IntegerType));
            getTempIlist.append(ifact.createReturn(IntegerType));
            mgen = new MethodGen(
                    Constants.ACC_PUBLIC, IntegerType,
                    new Type[] {}, new String[] {},
                    "getTemp" + fgen.getName(),
                    cname, getTempIlist, cpgen);
            mgen.setMaxLocals();
            mgen.setMaxStack();
            mgen.stripAttributes(true);
            cgen.addMethod(mgen.getMethod());
            
            // Create a 'set' accessor that sets the temp
            InstructionList setIlist = new InstructionList();
            setIlist.append(ifact.ALOAD_0);
            setIlist.append(ifact.ALOAD_1);
            setIlist.append(ifact.createPutField(cname, "temp_"+fgen.getName(), IntegerType));
            setIlist.append(ifact.RETURN);
            mgen = new MethodGen(
                    Constants.ACC_PUBLIC, Type.VOID,
                    new Type[] {IntegerType}, null,
                    "set" + fgen.getName(),
                    cname, setIlist, cpgen);
            mgen.setMaxLocals();
            mgen.setMaxStack();
            mgen.stripAttributes(true);
            cgen.addMethod(mgen.getMethod());
            
            // Create a commit that sets the state variable equal to the temp
            InstructionList commitIlist = new InstructionList();
            commitIlist.append(ifact.ALOAD_0);
            commitIlist.append(ifact.DUP);
            commitIlist.append(ifact.createGetField(cname, "temp_"+fgen.getName(), IntegerType));
            commitIlist.append(ifact.createPutField(cname, fgen.getName(), IntegerType));
            commitIlist.append(ifact.RETURN);
            mgen = new MethodGen(
                    Constants.ACC_PUBLIC, Type.VOID,
                    new Type[] {}, null,
                    "commit" + fgen.getName(),
                    cname, commitIlist, cpgen);
            mgen.setMaxLocals();
            mgen.setMaxStack();
            mgen.stripAttributes(true);
            cgen.addMethod(mgen.getMethod());
            
            // Generate field initialization code
            initIlist.append(ifact.ALOAD_0);
            initIlist.append(ifact.createNew((ObjectType)IntegerType));
            initIlist.append(ifact.DUP);
            initIlist.append(ifact.createConstant(state.start));
            initIlist.append(ifact.createInvoke("java.lang.Integer",
                    "<init>", Type.VOID, new Type[] {Type.INT},
                    Constants.INVOKESPECIAL));
            initIlist.append(ifact.createPutField(
                    cname, fgen.getName(), IntegerType));
            initIlist.append(ifact.ALOAD_0);
            initIlist.append(ifact.createNew((ObjectType)IntegerType));
            initIlist.append(ifact.DUP);
            initIlist.append(ifact.createConstant(state.start));
            initIlist.append(ifact.createInvoke("java.lang.Integer",
                    "<init>", Type.VOID, new Type[] {Type.INT},
                    Constants.INVOKESPECIAL));
            initIlist.append(ifact.createPutField(
                    cname, tempFgen.getName(), IntegerType));
        }

        // Add initialization code to all constructors
        Method[] methodArray = cgen.getMethods();
        for (int i = 0; i < methodArray.length; i++)
        {
            if (methodArray[i].getName().equals("<init>"))
            {
                mg = new MethodGen(methodArray[i], cname, cpgen);
                ilist = mg.getInstructionList();
                ilist.insert(initIlist.copy());
                mg.stripAttributes(true);
                mg.setMaxStack();
                mg.setMaxLocals();
                cgen.replaceMethod(cgen.containsMethod(mg.getName(), mg.getSignature()), mg.getMethod());
                ilist.dispose();
            }
        }
        
        initIlist.dispose();
    }

}
